/*-------------------------------------------------------------------------+
| Copyright 2023 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.ext.quality.ui.view.fx;

import static java.util.stream.Collectors.toList;
import static javafx.scene.paint.Color.BLUE;
import static javafx.scene.paint.Color.DARKGRAY;
import static javafx.scene.paint.Color.LIGHTGRAY;
import static org.fortiss.tooling.common.ui.javafx.style.FontStyle.BLACK_VERDANA_10PT;
import static org.fortiss.tooling.common.ui.javafx.style.FontStyle.BLACK_VERDANA_12PT;
import static org.fortiss.tooling.common.ui.javafx.style.FontStyle.BLACK_VERDANA_14PT;
import static org.fortiss.tooling.common.ui.javafx.style.FontStyle.BLACK_VERDANA_8PT;
import static org.fortiss.tooling.common.ui.javafx.style.LineStyle.SOLID_BLACK_1PT;
import static org.fortiss.tooling.kernel.ui.util.SelectionUtils.checkAndPickFirst;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getFirstParentWithType;

import java.text.DecimalFormat;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.ui.ISelectionListener;
import org.eclipse.ui.IWorkbenchPart;
import org.fortiss.tooling.common.ui.javafx.layout.CompositeFXControllerBase;
import org.fortiss.tooling.common.ui.javafx.style.FillStyle;
import org.fortiss.tooling.common.ui.javafx.style.FontStyle;
import org.fortiss.tooling.common.ui.javafx.style.LineStyle;
import org.fortiss.tooling.ext.quality.IMetricUpdateListener;
import org.fortiss.tooling.ext.quality.data.MetricData;
import org.fortiss.tooling.ext.quality.data.MetricKey;
import org.fortiss.tooling.ext.quality.data.MetricTreeNode;
import org.fortiss.tooling.ext.quality.service.IModelQualityService;
import org.fortiss.tooling.kernel.model.IProjectRootElement;
import org.fortiss.tooling.spiderchart.control.SpiderChartViewer;
import org.fortiss.tooling.spiderchart.model.DataSeries;
import org.fortiss.tooling.spiderchart.model.DoubleAxis;
import org.fortiss.tooling.spiderchart.model.SpiderChart;
import org.fortiss.tooling.spiderchart.style.AxisStyle;
import org.fortiss.tooling.spiderchart.style.ChartStyle;
import org.fortiss.tooling.spiderchart.style.DataSeriesStyle;
import org.fortiss.tooling.spiderchart.style.LegendStyle;

import javafx.application.Platform;
import javafx.fxml.FXML;
import javafx.scene.Node;
import javafx.scene.chart.BarChart;
import javafx.scene.chart.PieChart;
import javafx.scene.chart.ScatterChart;
import javafx.scene.chart.XYChart;
import javafx.scene.control.ChoiceBox;
import javafx.scene.control.Label;
import javafx.scene.control.SplitPane;
import javafx.scene.layout.BorderPane;
import javafx.scene.paint.Color;
import javafx.scene.text.Text;
import javafx.util.Pair;

/**
 * FX Controller for the Metrics view.
 *
 * @author groh
 */

@SuppressWarnings("unchecked")
public class ModelQualityFXController extends CompositeFXControllerBase<SplitPane, Node>
		implements ISelectionListener, IMetricUpdateListener {

	/**
	 * Selection of colors for the overlays in the spider chart. The length of this array also
	 * limits the depth of the {@link ModelQualityFXController#lastViewedElements} list.
	 */
	private static Color[] SPIDER_CHART_OVERLAY_COLORS = {Color.GREEN, Color.BLUE, Color.RED};

	/** {@link ChoiceBox} for selecting the metric which shall be displayed. */
	@FXML
	private ChoiceBox<MetricKey> childMetricChoiceBox;

	/** {@link PieChart} displaying the metrics. */
	@FXML
	private PieChart pieChart;

	/** The bottom {@link BorderPane} for the spider chart. */
	@FXML
	private BorderPane borderPane;

	/** {@link ChoiceBox} for selecting the metric displayed on the x axis of the scatter chart. */
	@FXML
	private ChoiceBox<MetricKey> xScatterMetricChoiceBox;

	/** {@link ChoiceBox} for selecting the metric displayed on the y axis of the scatter chart. */
	@FXML
	private ChoiceBox<MetricKey> yScatterMetricChoiceBox;

	/** A {@link ScatterChart} for displaying metrics. */
	@FXML
	private ScatterChart<Number, Number> scatterChart;

	/** A {@link BarChart} for displaying metrics. */
	@FXML
	private BarChart<String, Number> barChart;

	/** {@link Text} at the bottom of the view. */
	@FXML
	private Text bottomText;

	/** Time format when displaying updating time. */
	private final DateTimeFormatter timeFormat = DateTimeFormatter.ofPattern("HH:mm:ss");

	/**
	 * The last {@link EObject} which the user has clicked. Might be null when no or an invalid
	 * object is selected.
	 */
	private EObject lastSelectedElement;

	/** List of elements for which metrics were displayed. */
	private List<Pair<MetricTreeNode, Set<MetricKey>>> lastViewedElements = new ArrayList<>();

	/** {@inheritDoc} */
	@Override
	public String getFXMLLocation() {
		return "MetricsViewFx.fxml";
	}

	/** {@inheritDoc} */
	@Override
	public void initialize() {
		childMetricChoiceBox.getItems().addAll(MetricKey.values());
	}

	/** {@inheritDoc} */
	@Override
	public void selectionChanged(IWorkbenchPart part, ISelection selection) {

		// Invalidate last selection, so if no new selection is found the charts are cleared
		lastSelectedElement = null;

		EObject selected = checkAndPickFirst(selection, EObject.class);

		if(selected == null) {
			updateCharts("Nothing selected.");
			return;
		}

		IProjectRootElement currentRootElement =
				selected instanceof IProjectRootElement ? (IProjectRootElement)selected
						: getFirstParentWithType(selected, IProjectRootElement.class);

		if(currentRootElement == null) {
			updateCharts("Selection does not have a project root element.");
			return;
		}

		MetricData metricData = IModelQualityService.getInstance().getMetricData();
		var rootDataElement = metricData.getDataRootElementMap().get(currentRootElement);

		if(rootDataElement == null) {
			bottomText.setText(
					"Save model or select \"Export Metrics\" in project options to generate metrics.");
			updateCharts("No metrics available");
			return;
		}

		bottomText.setText("Metrics last updated: " +
				timeFormat.format(rootDataElement.getLastRefresh()) + ". Save model to refresh.");
		lastSelectedElement = selected;
		updateCharts("Selection does not have metric information.");
	}

	/** Triggered when the choice box changes. */
	public void onChoiceBoxChange() {
		updateCharts("Invalid selection or metrics not available.");
	}

	/** Triggered when the child metric choice box changes. */
	public void onChildMetricChoiceBoxChange() {
		updateCharts("Invalid selection or metrics not available.");
	}

	/** Triggered when one of the scatter choice boxes changes. */
	public void onScatterMetricChoiceBoxChange() {
		updateCharts("Invalid selection or metrics not available.");
	}

	/**
	 * Updates the chart in the metric view.
	 * 
	 * @param message
	 *            to display if metric is not available
	 */
	private void updateCharts(String message) {

		if(lastSelectedElement != null) {
			MetricData metricData = IModelQualityService.getInstance().getMetricData();
			MetricTreeNode node = metricData.getTreeNodeLookupTable().get(lastSelectedElement);

			if(node != null) {
				if(!node.getChildren().isEmpty()) {

					updatePieChart(node);
					updateScatterChart(node);
					updateBarChart(node);

					updateLastViewedElements(node);
					SpiderChartViewer viewer = buildSpiderChart();
					borderPane.setCenter(viewer.getViewerPane());

					// Return, otherwise the data will be cleared
					return;
				}
				message = "Selection does not have contained elements.";
			} else {
				message = "Selection does not have metric information.";
			}
		} else {
			// Nothing selected, so remove the bottom text
			bottomText.setText("");
		}
		borderPane.setCenter(new Label(message));
		pieChart.getData().clear();
	}

	/**
	 * Updates the {@link #pieChart} with data from the provided {@link MetricTreeNode}.
	 * 
	 * @param node
	 *            which provides the data for this chart
	 */
	private void updatePieChart(MetricTreeNode node) {
		MetricKey selectedMetric = childMetricChoiceBox.getValue();
		if(selectedMetric != null) {

			// Create PieChart and add data
			pieChart.getData().clear();
			pieChart.setLegendVisible(false);
			pieChart.setTitle("");
			DecimalFormat format = new DecimalFormat("0.#");

			for(var child : node.getChildren()) {
				Double value = child.getValueAsDouble(selectedMetric);
				if(value != null) {
					PieChart.Data slice =
							new PieChart.Data(child.getName() + ": " + format.format(value), value);
					pieChart.getData().add(slice);
				}
			}
		}
	}

	/**
	 * Updates the {@link #scatterChart} with data from the provided {@link MetricTreeNode}.
	 * 
	 * @param node
	 *            which provides the data for this chart
	 */
	private void updateScatterChart(MetricTreeNode node) {
		MetricKey xScatterKey = xScatterMetricChoiceBox.getValue();
		MetricKey yScatterKey = yScatterMetricChoiceBox.getValue();
		if(xScatterKey != null && yScatterKey != null) {

			scatterChart.getXAxis().setLabel(xScatterKey.toString());
			scatterChart.getYAxis().setLabel(yScatterKey.toString());
			scatterChart.getData().clear();

			for(var child : node.getChildren()) {
				Double xValue = child.getValueAsDouble(xScatterKey);
				Double yValue = child.getValueAsDouble(yScatterKey);

				if(xValue != null && yValue != null) {
					// Add data to the chart
					XYChart.Series<Number, Number> dataSeries = new XYChart.Series<>();
					dataSeries.setName(child.getName());
					dataSeries.getData().add(new XYChart.Data<>(xValue, yValue));
					scatterChart.getData().add(dataSeries);
				}
			}
		}
	}

	/**
	 * Updates the {@link #barChart} with data from the provided {@link MetricTreeNode}.
	 * 
	 * @param node
	 *            which provides the data for this chart
	 */
	private void updateBarChart(MetricTreeNode node) {
		MetricKey yScatterKey = yScatterMetricChoiceBox.getValue();
		if(yScatterKey != null) {

			barChart.getYAxis().setLabel(yScatterKey.toString());
			barChart.getData().clear();

			for(var child : node.getChildren()) {
				Double yValue = child.getValueAsDouble(yScatterKey);

				if(yValue != null) {
					// Add data to the chart
					XYChart.Series<String, Number> dataSeries = new XYChart.Series<>();
					dataSeries.setName(child.getName());
					dataSeries.getData().add(new XYChart.Data<>(child.getName(), yValue));
					barChart.getData().add(dataSeries);
				}
			}
		}
	}

	/**
	 * Creates a SpiderChart using the elements in
	 * {@link ModelQualityFXController#lastViewedElements}. A set of keys which are present in all
	 * elements in the list is created, and then these elements are overlaid.
	 */
	private SpiderChartViewer buildSpiderChart() {
		// Calculate the intersection of the keys, i.e. a set of keys which have a non-null value in
		// all elements of the lastViewedElements list
		// TODO (SB): Add check that lastViewedElements is non empty, and add a corresponding
		// comment
		// to the .get(0) statement
		Set<MetricKey> keysIntersection = lastViewedElements.get(0).getValue();
		for(var p : lastViewedElements) {
			keysIntersection.retainAll(p.getValue());
		}

		// Create DataSeries for all elements
		List<Pair<MetricTreeNode, DataSeries>> dataSeriesList =
				lastViewedElements.stream().map(pair -> {
					MetricTreeNode key = pair.getKey();
					String name = key.getName();
					return new Pair<>(key, (name == null) ? new DataSeries("No name found")
							: new DataSeries(name));
				}).collect(toList());

		// Create SpiderChart and set basic styling
		SpiderChart spiderChart = new SpiderChart();
		spiderChart.setLegendLabel("Legend");
		ChartStyle chartStyle = new ChartStyle(true, true, true);

		chartStyle.setUseIndividualAxisSegments(false);
		chartStyle.setTitleStyle(new FontStyle("Verdana", 14, BLUE.brighter()));

		// Iterate over all keys, and add data points
		for(MetricKey key : keysIntersection) {
			// Calculate maximumValue for the range of the axis
			Double maxmiumValue = lastViewedElements.stream()
					.mapToDouble(p -> p.getKey().getValueAsDouble(key)).max().getAsDouble();

			if(maxmiumValue == 0.0) {
				// Don't display metrics where all values are zero
				continue;
			}

			// Create Axis and apply Styling
			DoubleAxis doubleAxis = new DoubleAxis(key.toString(), 0.0, maxmiumValue);
			AxisStyle aStyle3Segs = new AxisStyle(SOLID_BLACK_1PT, BLACK_VERDANA_14PT, 3,
					BLACK_VERDANA_8PT, new DecimalFormat("#.##"));
			chartStyle.setAxisStyle(doubleAxis, aStyle3Segs);

			spiderChart.addAxis(doubleAxis);
			// Add the data point for this axis to all elements
			for(var pair : dataSeriesList) {
				pair.getValue().setPoint(doubleAxis, pair.getKey().getValueAsDouble(key));
			}
		}

		// Iterate over all dataSeries and add them to the spider chart
		int i = 0;
		for(var pair : dataSeriesList) {
			spiderChart.addData(pair.getValue());
			// Assign different colors to different data series
			// updateLastViewedElements() ensures that the are are at most three elements in
			// updateLastViewedElements, from which dataSeriesList is derived above.
			Color color = SPIDER_CHART_OVERLAY_COLORS[i];
			LineStyle olive1pt = new LineStyle(color);
			FillStyle oliveFill = new FillStyle(color, 0.25);
			DataSeriesStyle style = new DataSeriesStyle(olive1pt, oliveFill, true, true,
					BLACK_VERDANA_10PT, 7, new DecimalFormat("#.#"));
			chartStyle.setDataSeriesStyle(pair.getValue(), style);
			i++;
		}

		// Configure legend
		LegendStyle legendStyle = new LegendStyle(false, 5, BLACK_VERDANA_12PT);
		chartStyle.setLegendStyle(legendStyle);
		chartStyle.setAxisSegments(4);
		chartStyle.setBackgroundFillStyle(new FillStyle(LIGHTGRAY));
		chartStyle.setBackgroundLineStyle(new LineStyle(DARKGRAY));

		SpiderChartViewer viewer = new SpiderChartViewer(spiderChart, chartStyle);
		return viewer;
	}

	/** {@inheritDoc} */
	@Override
	public void metricUpdate(IProjectRootElement updatedElement) {
		Platform.runLater(() -> {
			if(lastSelectedElement == null) {
				// If nothing is selected, there is nothing to update
				return;
			}

			MetricData metricData = IModelQualityService.getInstance().getMetricData();
			var rootDataElement = metricData.getDataRootElementMap().get(updatedElement);
			if(rootDataElement != null) {
				bottomText.setText("Metrics last updated: " +
						timeFormat.format(rootDataElement.getLastRefresh()) +
						". Save model to refresh.");
			} else {
				bottomText.setText(
						"Save model or select \"Export Metrics\" in project options to generate metrics.");
			}
			updateCharts("Selection does not have metric information.");
		});
	}

	/**
	 * Updates the {@link ModelQualityFXController#lastViewedElements} variable. It will add the
	 * provided {@link MetricTreeNode} in the back of the list, and if the list is too long it will
	 * remove the first element.
	 * 
	 * @param node
	 *            which was selected
	 */
	private void updateLastViewedElements(MetricTreeNode node) {

		// Create set of keys which store values in the ndoe
		Set<MetricKey> keys = new HashSet<>();
		keys.addAll(node.getDoubleMetrics().keySet());
		keys.addAll(node.getIntegerMetrics().keySet());

		// Check if the node is already part of the list, and skip adding it if it is
		if(lastViewedElements.stream().allMatch(p -> p.getKey() != node)) {
			lastViewedElements.add(new Pair<>(node, keys));
		}

		// Limit the amount of elements in the list to the amount of colors
		if(lastViewedElements.size() > SPIDER_CHART_OVERLAY_COLORS.length) {
			lastViewedElements.remove(0);
		}
	}
}
